!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   pn_analyze_mod
!> @brief   run analyzing trajectories
!! @authors Norio Takase (NT)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../../../config.h"
#endif

module pn_analyze_mod

  use pn_option_str_mod
  use output_str_mod
  use input_str_mod
  use fileio_mod
  use string_mod
  use messages_mod
  use constants_mod

  implicit none
  private

  ! structures
  type s_data
    real(wp), allocatable :: v(:,:)
  end type s_data

  ! subroutines
  public  :: analyze
  private :: read_data
  private :: calc_pathcv
  private :: write_pathcv
  private :: get_replicate_name1

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    analyze
  !> @brief        run analyzing trajectories
  !! @authors      NT
  !! @param[in]    input    : input information
  !! @param[in]    output   : output information
  !! @param[in]    option   : option information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine analyze(input, output, option)

    ! formal arguments
    type(s_input),           intent(in)    :: input
    type(s_output),          intent(in)    :: output
    type(s_option),          intent(in)    :: option

    ! local variables
    type(s_data)             :: path, cv
    integer                  :: ireplica

    real(wp),     allocatable :: progress(:)
    real(wp),     allocatable :: distance(:)


    if (option%check_only) &
      return


    ! read path file
    !
    call read_data(input%pathfile, path)


    do ireplica = 1, option%nreplicas

      ! read cv files
      !
      call read_data(get_replicate_name1(input%cvfile, ireplica), cv)


      ! calculate path cv
      !
      call calc_pathcv(path, cv, progress, distance)


      ! write pathcv files
      !
      call write_pathcv(get_replicate_name1(output%pathcvfile, ireplica), &
                        progress, distance)

    end do

    return

  end subroutine analyze

  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_data(filename, data)

    ! formal arguments
    character(*),            intent(in)    :: filename
    type(s_data),            intent(inout) :: data

    ! local variables
    real(wp)                 :: dummy
    integer                  :: file, ndim, nstep, i, j
    character(1000)          :: line


    call open_file(file, filename, IOFileInput)

    read(file,'(a)') line
    ndim = split_num(line) - 1
    if (ndim == 0) &
      call error_msg('Read_Data> bad format. '//trim(filename))

    nstep = 1
    do while(.true.)
      read(file,*,end=10,err=10) line
      nstep = nstep + 1
    end do

10  rewind(file)
    write(MsgOut,'(a)')    'Read_Data> '
    write(MsgOut,'(a,a)')  '  file name = ', trim(filename)
    write(MsgOut,'(a,i0)') '    # of dimension : ', ndim
    write(MsgOut,'(a,i0)') '    # of steps     : ', nstep
    write(MsgOut,'(a)')    ''

    if (allocated(data%v)) &
      deallocate(data%v)
    allocate(data%v(nstep,ndim))
    do i = 1, nstep
      read(file,*) dummy, (data%v(i,j), j=1,ndim)
    end do

    call close_file(file)

    return

  end subroutine read_data

  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine calc_pathcv(path, cv, progress, distance)

    ! formal arguments
    type(s_data),            intent(in)    :: path
    type(s_data),            intent(in)    :: cv
    real(wp),                allocatable   :: progress(:)
    real(wp),                allocatable   :: distance(:)

    ! local variables
    real(wp)                 :: d_pathpoints, lambda
    integer                  :: ndim, nstep_p, nstep_c, istep_c, istep_p, idim

    real(wp),    allocatable :: v(:), log_m(:)
    real(wp),    allocatable :: dev(:,:), d(:,:), tmp(:,:)
    real(wp),    allocatable :: log_numerator(:), log_denominator(:)


    write(MsgOut,'(a)')    'Calc_Pathcv> '

    if (allocated(progress)) &
      deallocate(progress)

    if (allocated(distance)) &
      deallocate(distance)

    ndim    = size(path%v(1,:))
    nstep_p = size(path%v(:,1))
    nstep_c = size(cv%v(:,1))

    if (ndim /= size(cv%v(1,:))) &
      call error_msg('Calc_Pathcv> dimension of ref and data do not match.')

    allocate(progress(nstep_c))
    allocate(distance(nstep_c))

    allocate(v(ndim))

    ! determine lambda parameter
    d_pathpoints = 0
    do istep_p = 1, nstep_p-1
      v(1:ndim) = path%v(istep_p+1,1:ndim) - path%v(istep_p,1:ndim)
      v(1:ndim) = v(1:ndim) ** 2.0_wp
      d_pathpoints = d_pathpoints + sqrt(sum(v))
    end do
    d_pathpoints = d_pathpoints / real(nstep_p - 1,wp)
    lambda = 2.3_wp / (d_pathpoints ** 2.0_wp)

    ! calculate pathway CVs
    allocate(dev(nstep_c, ndim))
    allocate(d  (nstep_c, nstep_p))
    allocate(tmp(nstep_c, nstep_p))
    allocate(log_numerator  (nstep_c))
    allocate(log_denominator(nstep_c))

    do istep_p = 1, nstep_p
      do istep_c = 1, nstep_c
        dev(istep_c,1:ndim) = path%v(istep_p,1:ndim) - cv%v(istep_c,1:ndim)
        d(istep_c, istep_p) = sum(dev(istep_c,1:ndim)**2.0_wp)
      end do
    end do

    allocate(log_m(nstep_p))
    do istep_p = 1, nstep_p
      log_m(istep_p) = log(real(istep_p, wp))
    end do

    do istep_c = 1, nstep_c
      tmp(istep_c,1:nstep_p) = log_m(1:nstep_p) - lambda * d(istep_c,1:nstep_p)
    end do
    call logsumexp_array(tmp, log_numerator)

    do istep_c = 1, nstep_c
      tmp(istep_c,1:nstep_p) = - lambda * d(istep_c,1:nstep_p)
    end do
    call logsumexp_array(tmp, log_denominator)

    progress(1:nstep_c) = &
         exp(log_numerator(1:nstep_c) - log_denominator(1:nstep_c))
    
    distance(1:nstep_c) = &
         - (1.0_wp / lambda) * log_denominator(1:nstep_c)

    deallocate(log_m, &
               log_denominator, &
               log_numerator,   &
               tmp, d, dev, v)

    write(MsgOut,'(a)')    '  ..done'
    write(MsgOut,'(a)')    ''

    return

  end subroutine calc_pathcv

  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine write_pathcv(filename, progress, distance)

    ! formal arguments
    character(*),            intent(in)    :: filename
    real(wp),                intent(in)    :: progress(:)
    real(wp),                intent(in)    :: distance(:)

    ! local variables
    integer                  :: file, istep, nstep


    call open_file(file, filename, IOFileOutputNew)

    nstep = size(progress)

    do istep = 1, nstep
      write(file,'(i8,2f25.16)') istep, progress(istep), distance(istep)
    end do

    call close_file(file)

    return

  end subroutine write_pathcv

  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine logsumexp_array(x, s)

    ! formal arguments
    real(wp),                intent(in)    :: x(:,:)
    real(wp),                intent(inout) :: s(:)

    ! local variables
    real(wp)                 :: max_x
    integer                  :: nstep_c, nstep_p, istep_c

    real(wp),    allocatable :: v(:)


    nstep_p = size(x(1,:))
    nstep_c = size(x(:,1))

    allocate(v(nstep_p))

    do istep_c = 1, nstep_c

      max_x = maxval(x(istep_c,1:nstep_p))
      v(1:nstep_p) = exp(x(istep_c,1:nstep_p) - max_x)

      s(istep_c) = log(sum(v)) + max_x

    end do

    deallocate(v)

    return

  end subroutine logsumexp_array

  !======1=========2=========3=========4=========5=========6=========7=========8

  function get_replicate_name1(filename, no)

    ! return
    character(Maxfilename)   :: get_replicate_name1

    ! formal arguments
    character(*),            intent(in)    :: filename
    integer,                 intent(in)    :: no

    ! local variables
    integer                  :: bl, br


    bl = index(filename, '{', back=.true.)
    br = index(filename, '}', back=.true.)

    if (bl == 0 .or. br == 0 .or. bl > br) &
      call error_msg('Get_Replicate_Name1> Syntax error.')

    write(get_replicate_name1, '(a,i0,a)') &
         filename(:bl-1),no,filename(br+1:)

    return

  end function get_replicate_name1

end module pn_analyze_mod
