!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   sp_enefunc_mod
!> @brief   define potential energy functions in each domain
!! @authors Jaewoon Jung (JJ), Yuji Sugita (YS), Chigusa Kobayashi (CK)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module sp_enefunc_mod

  use sp_enefunc_gromacs_mod
  use sp_enefunc_amber_mod
  use sp_enefunc_charmm_mod
  use sp_enefunc_localres_mod
  use sp_enefunc_table_mod
  use sp_communicate_mod
  use sp_migration_mod
  use sp_energy_mod
  use sp_restraints_str_mod
  use sp_constraints_str_mod
  use sp_enefunc_str_mod
  use sp_energy_str_mod
  use sp_domain_str_mod
  use molecules_str_mod
  use fileio_localres_mod
  use fileio_grotop_mod
  use fileio_prmtop_mod
  use fileio_par_mod
  use timers_mod
  use messages_mod
  use mpi_parallel_mod
  use constants_mod
#ifdef MPI
  use mpi
#endif

  implicit none
  private

  ! subroutines
  public  :: define_enefunc
  public  :: define_enefunc_pio
  public  :: update_enefunc
  private :: setup_enefunc_bond_pio
  private :: setup_enefunc_bond_constraint_pio
  private :: setup_enefunc_angl_pio
  private :: setup_enefunc_dihe_pio
  private :: setup_enefunc_rb_dihe_pio
  private :: setup_enefunc_impr_pio
  private :: setup_enefunc_cmap_pio
  private :: setup_enefunc_nonb_pio
  private :: setup_enefunc_dispcorr
  private :: check_bonding

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    define_enefunc
  !> @brief        a driver subroutine for defining potential energy functions
  !! @authors      YS, JJ, CK
  !! @param[in]    ene_info    : ENERGY section control parameters information
  !! @param[in]    par         : CHARMM PAR information
  !! @param[in]    prmtop      : AMBER parameter topology information
  !! @param[in]    grotop      : GROMACS parameter topology information
  !! @param[in]    localres    : local restraint information
  !! @param[in]    molecule    : molecule information
  !! @param[in]    constraints : constraints information
  !! @param[inout] domain      : domain information
  !! @param[inout] enefunc     : energy potential functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine define_enefunc(ene_info, par, prmtop, grotop, localres, molecule, &
                            constraints, restraints, domain, enefunc)

    ! formal arguments
    type(s_ene_info),        intent(in)    :: ene_info
    type(s_par),             intent(in)    :: par
    type(s_prmtop),          intent(in)    :: prmtop
    type(s_grotop),          intent(in)    :: grotop
    type(s_localres),        intent(in)    :: localres
    type(s_molecule),        intent(in)    :: molecule
    type(s_constraints),     intent(inout) :: constraints
    type(s_restraints),      intent(in)    :: restraints
    type(s_domain),          intent(inout) :: domain
    type(s_enefunc),         intent(inout) :: enefunc


    enefunc%forcefield        = ene_info%forcefield
    enefunc%output_style      = ene_info%output_style
    enefunc%table%water_table = ene_info%table .and. &
                                (ene_info%water_model(1:4) == 'TIP3' .or. &
                                 ene_info%water_model(1:3) == 'WAT'  .or. &
                                 ene_info%water_model(1:3) == 'SOL') .and. &
                                (.not. ene_info%nonb_limiter)

    enefunc%switchdist        = ene_info%switchdist
    enefunc%cutoffdist        = ene_info%cutoffdist
    enefunc%pairlistdist      = ene_info%pairlistdist
    enefunc%dielec_const      = ene_info%dielec_const
    enefunc%force_switch      = ene_info%vdw_force_switch
    enefunc%vdw_shift         = ene_info%vdw_shift

    enefunc%pme_use           = ene_info%electrostatic == ElectrostaticPME
    enefunc%pme_alpha         = ene_info%pme_alpha
    enefunc%pme_ngrid_x       = ene_info%pme_ngrid_x
    enefunc%pme_ngrid_y       = ene_info%pme_ngrid_y
    enefunc%pme_ngrid_z       = ene_info%pme_ngrid_z
    enefunc%pme_nspline       = ene_info%pme_nspline
    enefunc%fft_scheme        = ene_info%fft_scheme
    enefunc%pme_max_spacing   = ene_info%pme_max_spacing
    enefunc%dispersion_corr   = ene_info%dispersion_corr
    enefunc%contact_check     = ene_info%contact_check
    enefunc%nonb_limiter      = ene_info%nonb_limiter
    enefunc%minimum_contact   = ene_info%minimum_contact


    ! charmm
    !
    if (par%num_bonds > 0) then

      call define_enefunc_charmm(ene_info, par, localres, molecule, &
                                 constraints, restraints, domain, enefunc)

    ! amber
    !
    else if (prmtop%num_atoms > 0) then

      call define_enefunc_amber (ene_info, prmtop, molecule, &
                                 constraints, restraints, domain, enefunc)

    ! gromacs
    !
    else if (grotop%num_atomtypes > 0) then

      call define_enefunc_gromacs(ene_info, grotop, molecule, &
                                 constraints, restraints, domain, enefunc)

    end if

    ! dispersion correction
    !
    call setup_enefunc_dispcorr(ene_info, domain, enefunc)

    ! bonding_checker
    !
    if (ene_info%contact_check) call check_bonding(enefunc, domain)


    return

  end subroutine define_enefunc

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    define_enefunc_pio
  !> @brief        a driver subroutine for defining potential energy functions
  !! @authors      JJ
  !! @param[in]    ene_info    : ENERGY section control parameters information
  !! @param[in]    localres    : local restraint information
  !! @param[inout] constraints : constraints information
  !! @param[inout] domain      : domain information
  !! @param[inout] enefunc     : energy potential functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine define_enefunc_pio(ene_info, localres, constraints, &
                                domain,enefunc)

    ! formal arguments
    type(s_ene_info),        intent(in)    :: ene_info
    type(s_localres),        intent(in)    :: localres
    type(s_constraints),     intent(inout) :: constraints
    type(s_domain),          intent(inout) :: domain
    type(s_enefunc),         intent(inout) :: enefunc

    ! local variables
    integer                  :: ncel, ncelb

    enefunc%forcefield        = ene_info%forcefield
    enefunc%output_style      = ene_info%output_style
    enefunc%table%water_table = ene_info%table .and. &
                                (ene_info%water_model(1:4) == 'TIP3' .or. &
                                 ene_info%water_model(1:3) == 'WAT'  .or. &
                                 ene_info%water_model(1:3) == 'SOL')

    enefunc%switchdist        = ene_info%switchdist
    enefunc%cutoffdist        = ene_info%cutoffdist
    enefunc%pairlistdist      = ene_info%pairlistdist
    enefunc%dielec_const      = ene_info%dielec_const
    enefunc%force_switch      = ene_info%vdw_force_switch
    enefunc%vdw_shift         = ene_info%vdw_shift

    enefunc%pme_use           = ene_info%electrostatic == ElectrostaticPME
    enefunc%pme_alpha         = ene_info%pme_alpha
    enefunc%pme_ngrid_x       = ene_info%pme_ngrid_x
    enefunc%pme_ngrid_y       = ene_info%pme_ngrid_y
    enefunc%pme_ngrid_z       = ene_info%pme_ngrid_z
    enefunc%pme_nspline       = ene_info%pme_nspline
    enefunc%pme_max_spacing   = ene_info%pme_max_spacing
    enefunc%fft_scheme        = ene_info%fft_scheme
    enefunc%dispersion_corr   = ene_info%dispersion_corr
    enefunc%contact_check     = ene_info%contact_check
    enefunc%nonb_limiter      = ene_info%nonb_limiter
    enefunc%minimum_contact   = ene_info%minimum_contact

    ! base
    !
    ncel  = domain%num_cell_local
    ncelb = domain%num_cell_local + domain%num_cell_boundary

    call alloc_enefunc  (enefunc, EneFuncBondCell, ncel, ncelb)

    if (.not. constraints%rigid_bond) then

      ! bond
      !
      call setup_enefunc_bond_pio(domain, enefunc)

      ! angle
      !
      call setup_enefunc_angl_pio(domain, enefunc)

    else

      ! bond
      !
      call setup_enefunc_bond_constraint_pio(domain, constraints, enefunc)

      ! angle
      !
      call setup_enefunc_angl_pio(domain, enefunc)

    end if

    ! dihedral
    !
    call setup_enefunc_dihe_pio(domain, enefunc)

    ! Ryckaert-Bellemans dihedral
    !
    call setup_enefunc_rb_dihe_pio(domain, enefunc)

    ! improper
    !
    call setup_enefunc_impr_pio(domain, enefunc)

    ! cmap
    !
    call setup_enefunc_cmap_pio(ene_info, domain, enefunc)

    ! nonbonded
    !
    call setup_enefunc_nonb_pio(ene_info, constraints, domain, enefunc)

    ! dispersion correction
    !
    call setup_enefunc_dispcorr(ene_info, domain, enefunc)

    ! lookup table
    !
    if(ene_info%table) &
    call setup_enefunc_table(ene_info, enefunc)

    ! restraint
    !
    call setup_enefunc_localres(localres, domain, enefunc)

    if (ene_info%contact_check) call check_bonding(enefunc, domain)

    ! write summary of energy function
    !
    if (main_rank) then
      write(MsgOut,'(A)') &
           'Define_Enefunc_Pio> Number of Interactions in Each Term'
      write(MsgOut,'(A20,I10,A20,I10)')                  &
           '  bond_ene        = ', enefunc%num_bond_all, &
           '  angle_ene       = ', enefunc%num_angl_all
      write(MsgOut,'(A20,I10,A20,I10)')                  &
           '  torsion_ene     = ', enefunc%num_dihe_all, &
           '  rb_torsion_ene  = ', enefunc%num_rb_dihe_all
      write(MsgOut,'(A20,I10,A20,I10)')                  &
           '  improper_ene    = ', enefunc%num_impr_all, &
           '  cmap_ene        = ', enefunc%num_cmap_all
      write(MsgOut,'(A20,I10,A20,I10)')                  &
           '  nb_exclusions   = ', enefunc%num_excl_all, &
           '  nb14_calc       = ', enefunc%num_nb14_all
      write(MsgOut,'(A)') ' '
    end if

    return

  end subroutine define_enefunc_pio

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    update_enefunc
  !> @brief        a driver subroutine for updating potential energy functions
  !! @authors      JJ
  !! @param[in]    table       : flag for table or not
  !! @param[inout] domain      : domain information
  !! @param[inout] comm        : communication information
  !! @param[inout] enefunc     : energy potential functions information
  !! @param[inout] constraints : constraints information [optional]
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine update_enefunc(table, domain, comm, enefunc, constraints)

    ! formal arguments
    logical,                       intent(in)    :: table
    type(s_domain),                intent(inout) :: domain
    type(s_comm),                  intent(inout) :: comm
    type(s_enefunc),               intent(inout) :: enefunc
    type(s_constraints), optional, intent(inout) :: constraints

    ! local variables
    logical                        :: first


    ! sending the bonding information to other domain
    !

    ! bond
    !
    call update_outgoing_enefunc_bond(domain, enefunc)

    ! angle
    !
    call update_outgoing_enefunc_angl(domain, enefunc)

    ! dihedral
    !
    call update_outgoing_enefunc_dihe(domain, enefunc)

    ! Ryckaert-Bellemans dihedral
    !
    call update_outgoing_enefunc_rb_dihe(domain, enefunc)

    ! improper dihedral
    !
    call update_outgoing_enefunc_impr(domain, enefunc)

    ! cmap
    !
    call update_outgoing_enefunc_cmap(domain, enefunc)

    ! restraint
    !
    call update_outgoing_enefunc_restraint(domain, enefunc)


    ! communicate neighbour domain
    !
    call communicate_bond(domain, comm, enefunc)


    ! bond
    !
    call update_incoming_enefunc_bond(domain, enefunc)

    ! angle
    !
    call update_incoming_enefunc_angl(domain, enefunc)

    ! dihedral
    !
    call update_incoming_enefunc_dihe(domain, enefunc)

    ! Ryckaert-Bellemans dihedral
    !
    call update_incoming_enefunc_rb_dihe(domain, enefunc)

    ! improper dihedral
    !
    call update_incoming_enefunc_impr(domain, enefunc)

    ! cmap
    !
    call update_incoming_enefunc_cmap(domain, enefunc)

    ! restraint
    !
    call update_incoming_enefunc_restraint(domain, enefunc)

    ! re-count nonbond exclusion list
    !

    first = .false.

    if (present(constraints)) then

      if (enefunc%local_restraint) then
        call count_nonb_excl_constraint_rest(first, table, constraints, &
                                             domain, enefunc)

      else
        call count_nonb_excl_constraint(first, table, constraints, &
                                             domain, enefunc)

      end if

    else

      if (enefunc%local_restraint) then
        if (table) then
          call count_nonb_excl_solute_rest(first, domain, enefunc)

        else
          call count_nonb_excl_rest(first, domain, enefunc)

        end if
      else
        if (table) then
          call count_nonb_excl_solute(first, domain, enefunc)

        else
          call count_nonb_excl(first, domain, enefunc)

        end if
      end if

    end if

    return

  end subroutine update_enefunc

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_bond_pio
  !> @brief        define BOND term for each cell in potential energy function
  !! @authors      NT
  !! @param[in]    domain  : domain information
  !! @param[inout] enefunc : potential energy functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_bond_pio(domain, enefunc)

    ! formal arguments
    type(s_domain),  target, intent(in)    :: domain
    type(s_enefunc), target, intent(inout) :: enefunc

    ! local variable
    integer                  :: i, found
    integer,         pointer :: bond(:)


    bond      => enefunc%num_bond

    found = 0
    do i = 1, domain%num_cell_local
      found = found + bond(i)
      if (bond(i) > MaxBond) &
        call error_msg('Setup_Enefunc_Bond_Pio> Too many bonds.')
    end do

#ifdef MPI
    call mpi_allreduce(found, enefunc%num_bond_all, 1, mpi_integer, &
                       mpi_sum, mpi_comm_country, ierror)
#else
    enefunc%num_bond_all = found
#endif

    return

  end subroutine setup_enefunc_bond_pio

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_bond_constraint_pio
  !> @brief        define BOND term for each cell in potential energy function
  !! @authors      NT
  !! @param[in]    domain      : domain information
  !! @param[in]    constraints : constraints information
  !! @param[inout] enefunc     : potential energy functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_bond_constraint_pio(domain, constraints, enefunc)

    ! formal arguments
    type(s_domain),      target, intent(in)    :: domain
    type(s_constraints), target, intent(inout) :: constraints
    type(s_enefunc),     target, intent(inout) :: enefunc

    ! local variable
    integer                  :: i, j, k, ih, found, ncel, connect
    integer,         pointer :: bond(:)
    integer,         pointer :: HGr_local(:,:)


    ! count # of bonds
    bond      => enefunc%num_bond

    found = 0
    do i = 1, domain%num_cell_local
      found = found + bond(i)
      if (bond(i) > MaxBond) &
        call error_msg('Setup_Enefunc_Bond_Constraint_Pio> Too many bonds.')
    end do

#ifdef MPI
    call mpi_allreduce(found, enefunc%num_bond_all, 1, mpi_integer, &
                       mpi_sum, mpi_comm_country, ierror)
#else
    enefunc%num_bond_all = found
#endif


    ! count # of constraints bonds
    HGr_local => constraints%HGr_local

    ncel    = domain%num_cell_local
    connect = constraints%connect

    found = 0

    do i = 1, ncel
      do j = 1, connect
        do k = 1, HGr_local(j,i)
          do ih = 1, j
            found = found + 1
          end do
        end do
      end do
    end do

#ifdef MPI
    call mpi_allreduce(found, constraints%num_bonds, 1, mpi_integer, &
                       mpi_sum, mpi_comm_country, ierror)
#else
    constraints%num_bonds = found
#endif


    return

  end subroutine setup_enefunc_bond_constraint_pio

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_angl_pio
  !> @brief        define ANGLE term for each cell in potential energy function
  !! @authors      NT
  !! @param[in]    domain  : domain information
  !! @param[inout] enefunc : potential energy functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_angl_pio(domain, enefunc)

    ! formal arguments
    type(s_domain),  target, intent(in)    :: domain
    type(s_enefunc), target, intent(inout) :: enefunc

    ! local variables
    integer                  :: i, found
    integer,         pointer :: angle(:)


    angle     => enefunc%num_angle

    found = 0
    do i = 1, domain%num_cell_local
      found = found + angle(i)
      if (angle(i) > MaxAngle) &
        call error_msg('Setup_Enefunc_Angl_Pio> Too many angles.')
    end do

#ifdef MPI
    call mpi_allreduce(found, enefunc%num_angl_all, 1, mpi_integer, &
                       mpi_sum, mpi_comm_country, ierror)
#else
    enefunc%num_angl_all = found
#endif

    return

  end subroutine setup_enefunc_angl_pio

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_dihe_pio
  !> @brief        define DIHEDRAL term in potential energy function
  !! @authors      NT
  !! @param[in]    domain  : domain information
  !! @param[inout] enefunc : potential energy functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_dihe_pio(domain, enefunc)

    ! formal arguments
    type(s_domain),   target, intent(in)    :: domain
    type(s_enefunc),  target, intent(inout) :: enefunc

    ! local variables
    integer                   :: i, found
    integer,          pointer :: dihedral(:)


    dihedral  => enefunc%num_dihedral

    found = 0
    do i = 1, domain%num_cell_local
      found = found + dihedral(i)
      if (dihedral(i) > MaxDihe) &
        call error_msg('Setup_Enefunc_Dihe_Pio> Too many dihedral angles.')
    end do

#ifdef MPI
    call mpi_allreduce(found, enefunc%num_dihe_all, 1, mpi_integer, &
                       mpi_sum, mpi_comm_country, ierror)
#else
    enefunc%num_dihe_all = found
#endif

    return

  end subroutine setup_enefunc_dihe_pio

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_rb_dihe_pio
  !> @brief        define Ryckaert-Bellemans DIHEDRAL term in potential energy
  !!               function
  !! @authors      NT
  !! @param[in]    domain  : domain information
  !! @param[inout] enefunc : potential energy functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_rb_dihe_pio(domain, enefunc)

    ! formal arguments
    type(s_domain),   target, intent(in)    :: domain
    type(s_enefunc),  target, intent(inout) :: enefunc

    ! local variables
    integer                   :: i, found
    integer,          pointer :: dihedral(:)


    dihedral  => enefunc%num_rb_dihedral

    found = 0
    do i = 1, domain%num_cell_local
      found = found + dihedral(i)
      if (dihedral(i) > MaxDihe) &
        call error_msg('Setup_Enefunc_RB_Dihe_Pio> Too many dihedral angles.')
    end do

#ifdef MPI
    call mpi_allreduce(found, enefunc%num_rb_dihe_all, 1, mpi_integer, &
                       mpi_sum, mpi_comm_country, ierror)
#else
    enefunc%num_rb_dihe_all = found
#endif

    return

  end subroutine setup_enefunc_rb_dihe_pio

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_impr_pio
  !> @brief        define IMPROPER term in potential energy function
  !! @authors      NT
  !! @param[in]    domain  : domain information
  !! @param[inout] enefunc : potential energy functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_impr_pio(domain, enefunc)

    ! formal variables
    type(s_domain),   target, intent(in)    :: domain
    type(s_enefunc),  target, intent(inout) :: enefunc

    ! local variables
    integer                   :: i, found
    integer,          pointer :: improper(:)


    improper  => enefunc%num_improper

    found = 0
    do i = 1, domain%num_cell_local
      found = found + improper(i)
      if (improper(i) > MaxImpr) &
        call error_msg( &
            'Setup_Enefunc_Impr_Pio> Too many improper dihedral angles')
    end do

#ifdef MPI
    call mpi_allreduce(found, enefunc%num_impr_all, 1, mpi_integer, mpi_sum, &
                       mpi_comm_country, ierror)
#else
    enefunc%num_impr_all = found
#endif

    return

  end subroutine setup_enefunc_impr_pio

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_cmap_pio
  !> @brief        define cmap term in potential energy function with DD
  !! @authors      NT
  !! @param[in]    ene_info : ENERGY section control parameters information
  !! @param[in]    domain   : domain information
  !! @param[inout] enefunc  : energy potential functions informationn
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_cmap_pio(ene_info, domain, enefunc)

    ! formal variables
    type(s_ene_info),        intent(in)    :: ene_info
    type(s_domain),  target, intent(in)    :: domain
    type(s_enefunc), target, intent(inout) :: enefunc

    ! local variables
    integer                  :: i, found


    found = 0
    do i = 1, domain%num_cell_local
      found = found + enefunc%num_cmap(i)
      if (enefunc%num_cmap(i) > MaxCmap) &
        call error_msg('Setup_Enefunc_Cmap_Pio> Too many cmaps.')
    end do

#ifdef MPI
    call mpi_allreduce(found, enefunc%num_cmap_all, 1, mpi_integer, mpi_sum, &
                       mpi_comm_country, ierror)
#else
    enefunc%num_cmap_all = found
#endif

    return

  end subroutine setup_enefunc_cmap_pio

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_nonb_pio
  !> @brief        define NON-BOND term in potential energy function
  !! @authors      NT
  !! @param[in]    ene_info    : ENERGY section control parameters information
  !! @param[in]    molecule    : molecule information
  !! @param[in]    constraints : constraints information
  !! @param[inout] domain      : domain information
  !! @param[inout] enefunc     : energy potential functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_nonb_pio(ene_info, constraints, domain, enefunc)

    ! formal arguments
    type(s_ene_info),        intent(in)    :: ene_info
    type(s_constraints),     intent(in)    :: constraints
    type(s_domain),          intent(inout) :: domain
    type(s_enefunc),         intent(inout) :: enefunc

    ! local variables
    integer                  :: ncel


    ! treatment for 1-2, 1-3, 1-4 interactions
    !
    ncel   = domain%num_cell_local

    call alloc_enefunc(enefunc, EneFuncNonb,     ncel, maxcell)
    call alloc_enefunc(enefunc, EneFuncNonbList, ncel, maxcell)

    if (enefunc%table%water_table) then

      if (constraints%rigid_bond) then
        call count_nonb_excl_constraint(.true., .true., constraints, &
                                        domain, enefunc)

      else
        call count_nonb_excl_solute(.true., domain, enefunc)

      end if

    else

      if (constraints%rigid_bond) then
        call count_nonb_excl_constraint(.true., .false., constraints, &
                                        domain, enefunc)

      else
        call count_nonb_excl(.true., domain, enefunc)

      end if

    end if

    return

  end subroutine setup_enefunc_nonb_pio

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_dispcorr
  !> @brief        define dispersion correction term
  !! @authors      CK
  !! @param[in]    ene_info : ENERGY section control parameters information
  !! @param[inout] domain   : domain information
  !! @param[inout] enefunc  : energy potential functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  subroutine setup_enefunc_dispcorr(ene_info, domain, enefunc)

    ! formal arguments
    type(s_ene_info),        intent(in)    :: ene_info
    type(s_domain),  target, intent(inout) :: domain
    type(s_enefunc), target, intent(inout) :: enefunc

    ! local variables
    integer                  :: i, j, iatmcls, ntypes
    integer                  :: icel1, icel2, icel3, icel4
    integer                  :: i1, i2, i3, i4
    integer                  :: num_all_atoms, natom2, nexpair
    real(dp)                 :: lj6_tot, lj6_diff, lj6_ex
    real(dp)                 :: factor, rpair
    real(dp)                 :: diff_cs, diff_cs2, diff_cs3, diff_cs4
    real(dp)                 :: cutoff , cutoff2, cutoff3, cutoff4
    real(dp)                 :: cutoff5, cutoff6, cutoff7, cutoff8
    real(dp)                 :: cutoff14
    real(dp)                 :: inv_cutoff3, inv_cutoff6, inv_cutoff12
    real(dp)                 :: switchdist , switchdist2, switchdist3
    real(dp)                 :: switchdist4, switchdist5
    real(dp)                 :: switchdist6, switchdist7, switchdist8
    real(dp)                 :: shift_a, shift_b, shift_c
    real(dp)                 :: vswitch, eswitch, vlong

    integer,         pointer :: id_g2l(:,:)
    integer,         pointer :: bondlist(:,:,:),anglelist(:,:,:)
    integer,         pointer :: dihelist(:,:,:),rb_dihelist(:,:,:)
    integer,         pointer :: atmcls(:,:),imprlist(:,:,:)
    integer,     allocatable :: atype(:)


    if (ene_info%dispersion_corr == Disp_corr_NONE) return

    bondlist    => enefunc%bond_list
    anglelist   => enefunc%angle_list
    dihelist    => enefunc%dihe_list
    rb_dihelist => enefunc%rb_dihe_list
    imprlist    => enefunc%impr_list
    atmcls      => domain%atom_cls_no
    id_g2l      => domain%id_g2l

    ntypes = enefunc%num_atom_cls
    allocate(atype(1:ntypes))

    atype(1:ntypes) = 0
    num_all_atoms   = 0

    do i = 1, domain%num_cell_local
      do j = 1, domain%num_atom(i)
        iatmcls = atmcls(j,i)
        atype(iatmcls) = atype(iatmcls)+1
      end do
      num_all_atoms = num_all_atoms + domain%num_atom(i)
    end do

#ifdef MPI
    call mpi_allreduce(mpi_in_place, atype, ntypes, mpi_integer, &
                       mpi_sum, mpi_comm_country, ierror)
#endif

    lj6_tot = 0.0_dp
    do i = 1, ntypes
      do j = 1, ntypes
        lj6_tot = lj6_tot + enefunc%nonb_lj6(i,j)*atype(i)*atype(j)
      end do
    end do
    deallocate(atype)

    cutoff       = enefunc%cutoffdist
    cutoff2      = cutoff*cutoff
    cutoff3      = cutoff2*cutoff
    inv_cutoff3  = 1.0_dp/cutoff3

    eswitch = 0.0_dp
    vswitch = 0.0_dp
    vlong   = inv_cutoff3/3.0_dp

    if (enefunc%forcefield == ForcefieldAMBER ) then

      factor       = 2.0_dp*PI*lj6_tot
      enefunc%dispersion_energy = -factor*vlong
      enefunc%dispersion_virial = -2.0_dp*factor*vlong

    else if (enefunc%forcefield == ForcefieldGROAMBER .or.  &
             enefunc%forcefield == ForcefieldGROMARTINI) then
      !
      ! remove exclusion
      !
      lj6_ex = 0.0_dp
      nexpair = 0
      do i = 1, domain%num_cell_local
        ! self
        do j = 1, domain%num_atom(i)
          iatmcls = atmcls(j,i)
          lj6_ex  = lj6_ex + enefunc%nonb_lj6(iatmcls,iatmcls)
        end do

        ! bonds
        do j = 1, enefunc%num_bond(i)
          icel1 = id_g2l(1,bondlist(1,j,i))
          i1    = id_g2l(2,bondlist(1,j,i))
          icel2 = id_g2l(1,bondlist(2,j,i))
          i2    = id_g2l(2,bondlist(2,j,i))
          lj6_ex= lj6_ex + enefunc%nb14_lj6(atmcls(i1,icel1),atmcls(i2,icel2))
        end do

        ! angles
        do j = 1, enefunc%num_angle(i)
          icel1 = id_g2l(1,anglelist(1,j,i))
          i1    = id_g2l(2,anglelist(1,j,i))
          icel3 = id_g2l(1,anglelist(3,j,i))
          i3    = id_g2l(2,anglelist(3,j,i))
          lj6_ex= lj6_ex + enefunc%nb14_lj6(atmcls(i1,icel1),atmcls(i3,icel3))
        end do

        ! dihedral
        do j = 1, enefunc%num_dihedral(i)
          icel1 = id_g2l(1,dihelist(1,j,i))
          i1    = id_g2l(2,dihelist(1,j,i))
          icel4 = id_g2l(1,dihelist(4,j,i))
          i4    = id_g2l(2,dihelist(4,j,i))
          lj6_ex= lj6_ex + enefunc%nb14_lj6(atmcls(i1,icel1),atmcls(i4,icel4))
        end do

        ! RB dihedral
        do j = 1, enefunc%num_rb_dihedral(i)
          icel1 = id_g2l(1,rb_dihelist(1,j,i))
          i1    = id_g2l(2,rb_dihelist(1,j,i))
          icel4 = id_g2l(1,rb_dihelist(4,j,i))
          i4    = id_g2l(2,rb_dihelist(4,j,i))
          lj6_ex= lj6_ex + enefunc%nb14_lj6(atmcls(i1,icel1),atmcls(i4,icel4))
        end do

        ! improper
        do j = 1, enefunc%num_improper(i)
          icel1 = id_g2l(1,imprlist(1,j,i))
          i1    = id_g2l(2,imprlist(1,j,i))
          icel4 = id_g2l(1,imprlist(4,j,i))
          i4    = id_g2l(2,imprlist(4,j,i))
          lj6_ex= lj6_ex + enefunc%nb14_lj6(atmcls(i1,icel1),atmcls(i4,icel4))
        end do

        nexpair = nexpair + domain%num_atom(i)        &
                          + enefunc%num_bond(i)        &
                          + enefunc%num_angle(i)       &
                          + enefunc%num_dihedral(i)    &
                          + enefunc%num_rb_dihedral(i) &
                          + enefunc%num_improper(i)
      end do
#ifdef MPI
      call mpi_allreduce(mpi_in_place, num_all_atoms, 1, mpi_integer, &
                         mpi_sum, mpi_comm_country, ierror)
      call mpi_allreduce(mpi_in_place, nexpair, 1, mpi_integer, &
                         mpi_sum, mpi_comm_country, ierror)
      call mpi_allreduce(mpi_in_place, lj6_ex, 1, mpi_real8, &
                         mpi_sum, mpi_comm_country, ierror)
#endif
      lj6_diff = (lj6_tot - lj6_ex)

      natom2 = num_all_atoms*num_all_atoms
      rpair  = real(natom2/(natom2-nexpair),dp)
      factor       = 2.0_dp*PI*rpair*lj6_diff

      switchdist   = enefunc%switchdist
      diff_cs      = (cutoff - switchdist)

      if (diff_cs > EPS) then

        if (enefunc%vdw_shift) then
          cutoff4      = cutoff3*cutoff
          cutoff5      = cutoff4*cutoff
          cutoff6      = cutoff5*cutoff
          cutoff7      = cutoff6*cutoff
          cutoff8      = cutoff7*cutoff
          cutoff14     = cutoff7*cutoff7
          inv_cutoff6  = inv_cutoff3*inv_cutoff3
          inv_cutoff12 = inv_cutoff6*inv_cutoff6
  
          diff_cs2     = diff_cs*diff_cs
          diff_cs3     = diff_cs2*diff_cs
          diff_cs4     = diff_cs3*diff_cs
  
          switchdist2  = switchdist*switchdist
          switchdist3  = switchdist2*switchdist
          switchdist4  = switchdist3*switchdist
          switchdist5  = switchdist4*switchdist
          switchdist6  = switchdist5*switchdist
          switchdist7  = switchdist6*switchdist
          switchdist8  = switchdist7*switchdist
  
          ! LJ6
          !
          shift_a = -(10.0_dp*cutoff - 7.0_dp*switchdist)/(cutoff8*diff_cs2)
          shift_b =  ( 9.0_dp*cutoff - 7.0_dp*switchdist)/(cutoff8*diff_cs3)
  
          shift_c = inv_cutoff6 - 2.0_dp * shift_a * diff_cs3  &
                    - 1.5_dp * shift_b * diff_cs4
  
          eswitch = -2.0_dp * shift_a * ((1.0_dp/6.0_dp)*cutoff6                &
                                        -(3.0_dp/5.0_dp)*cutoff5*switchdist     &
                                        +(3.0_dp/4.0_dp)*cutoff4*switchdist2    &
                                        -(1.0_dp/3.0_dp)*cutoff3*switchdist3    &
                                        +(1.0_dp/6.0e1_dp)*switchdist6)         &
                    -1.5_dp * shift_b * ((1.0_dp/7.0_dp)*cutoff7                &
                                        -(2.0_dp/3.0_dp)*cutoff6*switchdist     &
                                        +(6.0_dp/5.0_dp)*cutoff5*switchdist2    &
                                        -                cutoff4*switchdist3    &
                                        +(1.0_dp/3.0_dp)*cutoff3*switchdist4    &
                                        -(1.0_dp/1.05e2_dp)*switchdist7)        &
                    -(1.0_dp/3.0_dp) * shift_c * (cutoff3)
    
          ! LJ12
          !
          shift_a = -(16.0_dp*cutoff - 13.0_dp*switchdist)/(cutoff14*diff_cs2)
          shift_b =  (15.0_dp*cutoff - 13.0_dp*switchdist)/(cutoff14*diff_cs3)
          shift_c = inv_cutoff12 - 2.0_dp * shift_a * diff_cs3  &
                    - 1.5_dp * shift_b * diff_cs4
  
 
          shift_a = -(10.0_dp*cutoff - 7.0_dp*switchdist)/(cutoff8*diff_cs2)
          shift_b =  ( 9.0_dp*cutoff - 7.0_dp*switchdist)/(cutoff8*diff_cs3)
 
          vswitch = shift_a * ( (1.0_dp/6.0_dp)*cutoff6                         &
                               -(2.0_dp/5.0_dp)*cutoff5*switchdist              &
                               +(1.0_dp/4.0_dp)*cutoff4*switchdist2             &
                               -(1.0_dp/6.0e1_dp)*switchdist6)                  &
                   +shift_b * ( (1.0_dp/7.0_dp)*cutoff7                         &
                               -(1.0_dp/2.0_dp)*cutoff6*switchdist              &
                               +(3.0_dp/5.0_dp)*cutoff5*switchdist2             &
                               -(1.0_dp/4.0_dp)*cutoff4*switchdist3             &
                               +(1.0_dp/1.4e2_dp)*switchdist7)
        enefunc%dispersion_energy = factor*(eswitch-vlong)
        enefunc%dispersion_virial = -2.0_dp*factor*(-vswitch+vlong)

        else

          eswitch = enefunc%eswitch
          vswitch = enefunc%vswitch
          enefunc%dispersion_energy = factor*(eswitch-vlong)
          enefunc%dispersion_virial = -factor*(vswitch+vlong)

        end if

      else 

        enefunc%dispersion_energy = factor*(eswitch-vlong)
        enefunc%dispersion_virial = -2.0_dp*factor*(-vswitch+vlong)

      end if

    else
      call error_msg('Setup_Enefunc_DispCorr> This force field is not allowed')
    end if

!   enefunc%dispersion_energy = factor*(eswitch-vlong)
!   enefunc%dispersion_virial = -2.0_wp*factor*(-vswitch+vlong)

  end subroutine setup_enefunc_dispcorr

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    check_bonding
  !> @brief        check bonds
  !! @authors      CK
  !! @param[in]    enefunc  : potential energy functions information
  !! @param[in]    domain   : domain information
  ! 
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine check_bonding(enefunc, domain)

    ! formal arguments
    type(s_enefunc), target, intent(in)    :: enefunc
    type(s_domain),  target, intent(in)    :: domain

    ! local variables
    real(wp)                 :: d12(1:3), r12, r_dif
    integer                  :: i, j, ix, icel1, icel2, i1, i2
    integer                  :: id, my_id, omp_get_thread_num
    real(wp), parameter      :: maxdistance = 0.5_wp

    real(wp),        pointer :: r0(:,:)
    integer,         pointer :: nbond(:), bondlist(:,:,:)
    integer,         pointer :: ncell_local
    integer,         pointer :: id_g2l(:,:)
    real(dp),        pointer :: coord(:,:,:)

    ncell_local => domain%num_cell_local
    id_g2l      => domain%id_g2l
    coord       => domain%coord

    nbond       => enefunc%num_bond
    bondlist    => enefunc%bond_list
    r0          => enefunc%bond_dist_min

    !$omp parallel default(shared)                                     &
    !$omp private(id, i, j, ix, icel1, i1, icel2, i2, d12, r12, r_dif, &
    !$omp         my_id)
    !
#ifdef OMP
    id  = omp_get_thread_num()
#else
    id  = 0
#endif
    my_id = id

    do i = my_id+1, ncell_local, nthread

      do ix = 1, nbond(i)

        icel1 = id_g2l(1,bondlist(1,ix,i))
        i1    = id_g2l(2,bondlist(1,ix,i))
        icel2 = id_g2l(1,bondlist(2,ix,i))
        i2    = id_g2l(2,bondlist(2,ix,i))

        d12(1:3) = coord(1:3,i1,icel1) - coord(1:3,i2,icel2)
        r12   = sqrt( d12(1)*d12(1) + d12(2)*d12(2) + d12(3)*d12(3) )
        r_dif = r12 - r0(ix,i)
        if (r_dif .gt. maxdistance) &
           write(MsgOut,'(A,I10,I10,F10.5)') &
          'WARNING: too long bond:',bondlist(1,ix,i),bondlist(2,ix,i),r12
        if (r_dif .lt. -maxdistance) &
           write(MsgOut,'(A,I10,I10,F10.5)') &
          'WARNING: too short bond:',bondlist(1,ix,i),bondlist(2,ix,i),r12

      end do

    end do

    !$omp end parallel 

    return

  end subroutine check_bonding

end module sp_enefunc_mod
